"""
Configuration tests.
"""

import os
from ditz.config import config


def test_config():
    "Test configuration files"

    config.write_file("test-config.ini")


if __name__ == "__main__":
    test_config()



# -*- coding: utf-8 -*-

"""
Ditz command handler.
"""

import os
import re
import sys

from cmd import Cmd
from database import DitzDB
from objects import Project, Config, find_config
from util import print_columns, error, default_name, default_email, DitzError

from flags import (STATUS, IN_PROGRESS, PAUSED, CLOSED, DISPOSITION,
                   FIXED, WONTFIX, REORG, BUGFIX, FEATURE, TASK,
                   UNRELEASED)


class DitzCmd(Cmd):
    prompt = "Ditz: "

    def __init__(self, path=".", **kwargs):
        Cmd.__init__(self)

        #: Whether running interactively.
        self.interactive = kwargs.pop("interactive", False)

        #: Whether to skip asking for comments.
        self.nocomments = kwargs.pop("nocomment", False)

        #: Comment string to use in non-interactive mode.
        self.comment = kwargs.pop("comment", None)

        #: Database (read when required).
        self._db = None

        #: Arguments for reading database.
        self._dbargs = (os.path.realpath(path), kwargs)

        #: Last issue name mentioned.
        self.last_issuename = None

        #: Last release name mentioned.
        self.last_releasename = None

        #: list of available commands.
        self.commands = set()

        # Set up list of known commands.
        for name in self.get_names():
            if name.startswith("do_"):
                self.commands.add(name[3:])

        # If interactive, set things up right now.
        if self.interactive:
            self.initdb()

    @property
    def db(self):
        self.initdb()
        return self._db

    def initdb(self):
        """
        Read or create the Ditz database, if not already done.
        """

        if not self._db:
            path, kwargs = self._dbargs
            if self.interactive:
                if find_config(path):
                    print "Reading Ditz database in", path
                    print
                else:
                    print "Setting up Ditz database in", path
                    print
                    if not self.setup():
                        sys.exit()

            self._db = DitzDB.read(path, **kwargs)

        return self._db

    def setup(self):
        """
        Set up a database interactively.
        """

        config = self.getconfig()
        if config:
            name = config.name
            email = config.email
            issuedir = config.issue_dir
        else:
            return

        path = os.path.join(issuedir, Project.filename)
        if os.path.exists(path):
            config.write()
            print
            print "Configuration written"
        else:
            path, kwargs = self._dbargs
            project = self.getline("Project name", os.path.basename(path))
            if not project:
                return

            db = DitzDB(project, name, email, issuedir, path, **kwargs)
            db.write()

            print
            print "Ditz database created in", path

        return True

    # Command methods.

    def do_add(self, arg):
        "add -- Add an issue"

        self.initdb()

        title = self.getline("Title: ")
        if not title:
            return

        desc = self.gettext("Description")
        if desc is None:
            return

        types = {'b': BUGFIX, 'f': FEATURE, 't': TASK}

        while True:
            reply = self.getline("Is this a (b)ugfix, (f)eature or (t)ask? ")
            if reply and reply[0] in types:
                issuetype = types[reply[0]]
                break

        choices = self.db.components
        if len(choices) >= 2:
            comp = self.getchoice("component", choices)
            if not comp:
                return
        else:
            comp = None

        release = None
        releases = self.db.project.releases
        choices = [r.name for r in releases if r.status == UNRELEASED]
        if choices and self.getyesno("Assign to a release now?"):
            if len(choices) > 1:
                release = self.getchoice("release", choices)
            else:
                release = choices[0]
                print "Assigning to release", release

            if not release:
                return

        default = self.db.config.username
        prompt = 'Issue creator (enter for "%s"): ' % default
        reporter = self.getline(prompt)
        comment = self.getcomment()
        if comment is None:
            return

        issue = self.db.add_issue(title, desc, type=issuetype,
                                  reporter=reporter, release=release,
                                  component=comp, comment=comment)

        print "Added issue", self.db.issue_name(issue)
        return True

    def do_add_component(self, arg):
        "add-component [name] -- Add a component"

        self.initdb()

        name = self.getarg(arg, 1) or self.getline("Name: ")
        if not name:
            return

        self.db.add_component(name)
        print "Added component", name
        return True

    def do_add_reference(self, arg):
        "add-reference <issue> -- Add a reference to an issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        name = self.db.issue_name(issue)
        print "Adding a reference to %s: %s" % (name, issue.title)
        text = self.getline("Reference: ")
        if not text:
            return

        comment = self.getcomment()
        if comment is None:
            return

        self.db.add_reference(issue, text, comment=comment)
        print "Added reference to", name
        return True

    def do_add_release(self, arg):
        "add-release [name] -- Add a release"

        self.initdb()

        name = self.getarg(arg, 1) or self.getline("Name: ")
        if not name:
            return

        print "Adding release", name
        comment = self.getcomment()
        if comment is None:
            return

        self.db.add_release(name, comment=comment)
        print "Added release", name
        return True

    def do_archive(self, arg):
        "archive <release> [dir] -- Archive a release"

        self.initdb()

        name = self.getrelease(arg, 1)
        if not name:
            return

        path = self.getarg(arg, 2)
        if not path:
            path = "ditz-archive-%s" % name

        self.db.archive_release(name, path)
        print "Archived to", path

    def do_assign(self, arg):
        "assign <issue> [release] -- Assign an issue to a release"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        name = self.db.issue_name(issue)
        if issue.release:
            print "Issue %s currently assigned to release %s" % \
                  (name, issue.release)
        else:
            print "Issue %s not currently assigned to any release" % name

        release = self.getarg(arg, 2)

        if not release:
            releases = self.db.project.releases
            choices = [r.name for r in releases if r.status == UNRELEASED]

            if issue.release and issue.release in choices:
                choices.remove(issue.release)

            if not choices:
                error("no other release available")
                return

            if len(choices) > 1:
                release = self.getchoice("release", choices)
            else:
                release = choices[0]

        if not release:
            return

        if release == issue.release:
            error("already assigned to release %s" % release)
            return

        if not self.db.get_release(release):
            error("no release with name %s" % release)
            return

        print "Assigning to release", release
        comment = self.getcomment()
        if comment is None:
            return

        self.db.set_release(issue, release, comment=comment)
        print "Assigned", name, "to", release

        return True

    def do_changelog(self, arg):
        "changelog <release> -- Generate a changelog for a release"

        self.initdb()

        name = self.getrelease(arg, 1)
        if not name:
            return

        print self.db.show_changelog(name)
        return True

    def do_close(self, arg):
        "close <open_issue> -- Close an issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        name = self.db.issue_name(issue)
        print "Closing issue %s: %s" % (name, issue.title)

        disp = DISPOSITION
        revdisp = {disp[x]: x for x in disp}
        choices = (disp[FIXED], disp[WONTFIX], disp[REORG])
        choice = self.getchoice("disposition", choices)
        if not choice:
            return

        disp = revdisp[choice]
        comment = self.getcomment()
        if comment is None:
            return

        self.db.set_status(issue, CLOSED, disposition=disp, comment=comment)
        print "Closed issue", name, "with disposition", choice
        return True

    def do_comment(self, arg):
        "comment <issue> -- Comment on an issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        name = self.db.issue_name(issue)
        print "Commenting on %s: %s" % (name, issue.title)

        comment = self.getcomment()

        if comment:
            self.db.add_comment(issue, comment)
            print "Comments recorded for", name
        else:
            print "Empty comment, aborted"

        return True

    def do_drop(self, arg):
        "drop <issue> -- Drop an issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        name = self.db.issue_name(issue)
        self.db.drop_issue(issue)
        print "Dropped %s.  Other issue names may have changed." % name
        return True

    def do_edit(self, arg):
        "edit <issue> -- Edit an issue"

        self.initdb()
        return self.unimplemented()

    def do_grep(self, arg):
        "grep <string> -- Show issues matching a string or regular expression"

        self.initdb()

        string = self.getarg(arg, 1)
        if not string:
            error("no regexp specified")
            return

        try:
            regexp = re.compile(string)
        except re.error as msg:
            error("invalid regexp: %s" % str(msg))
            return

        print self.db.show_grep(regexp) or "No matching issues"
        return True

    def do_html(self, arg):
        "html [dir] -- Generate HTML status pages"

        self.initdb()
        path = self.getarg(arg, 1) or "html"
        self.db.export_html(path)

        return True

    def do_init(self, arg):
        "init -- Initialize the issue database for a new project"

        path = find_config(self._dbargs[0])
        if path:
            dirname = os.path.split(path)[0]
            error("Ditz database already exists in %s" % dirname)
            return

        self.setup()

    def do_log(self, arg):
        "log [count] -- Show recent activity"

        self.initdb()

        count = self.getint(arg, 1)
        for line in self.db.log_events(count=count, verbose=True):
            print line

        return True

    def do_reconfigure(self, arg):
        "reconfigure -- Rerun configuration script"

        db = self.db
        config = self.getconfig(db.config)
        if config:
            db.config = config
            db.config.write(db.path)
            print "Configuration written"

    def do_release(self, arg):
        "release <unreleased_release> -- Release a release"

        self.initdb()

        name = self.getrelease(arg, 1)
        if not name:
            return

        comment = self.getcomment()
        if comment is None:
            return

        self.db.release_release(name, comment)
        print "Released", name
        return True

    def do_releases(self, arg):
        "releases -- Show releases"

        self.initdb()

        print self.db.show_releases() or "No releases"
        return True

    def do_set_component(self, arg):
        "set-component <issue> [component] -- Set an issue's component"

        self.initdb()

        issue = self.getissue(arg, 1)
        comp = self.getarg(arg, 2)
        if not issue:
            return

        name = self.db.issue_name(issue)
        print "Changing the component of issue %s: %s" % (name, issue.title)

        choices = self.db.components
        if len(choices) < 2:
            error("this project does not use multiple components")
            return

        if comp and comp not in choices:
            error("unknown component: %s" % comp)
            return

        if not comp:
            if issue.component in choices:
                choices.remove(issue.component)

            comp = self.getchoice("component", choices)
            if not comp:
                return

        comment = self.getcomment()
        if comment is None:
            return

        self.db.set_component(issue, comp, comment)
        newname = self.db.issue_name(issue)

        print "Issue %s is now %s.  Other issue names " \
              "may have changed" % (name, newname)

        return True

    def do_shortlog(self, arg):
        "shortlog [count] -- Show recent activity (short form)"

        self.initdb()

        count = self.getint(arg, 1)
        for line in self.db.log_events(count=count):
            print line

        return True

    def do_show(self, arg):
        "show <issue> -- Describe a single issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if issue:
            print self.db.show_issue(issue)

        return True

    def do_start(self, arg):
        "start <unstarted_issue> -- Start work on an issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        if issue.status == IN_PROGRESS:
            error("already marked as " + STATUS[issue.status])
            return

        name = self.db.issue_name(issue)
        print "Starting work on %s: %s" % (name, issue.title)
        comment = self.getcomment()
        if comment is None:
            return

        self.db.set_status(issue, IN_PROGRESS, comment=comment)
        print "Recorded start of work for", name
        return True

    def do_status(self, arg):
        "status [release] -- Show project status"

        self.initdb()

        release = self.getrelease(arg, 1, optional=True)
        print self.db.show_status(release)

    def do_stop(self, arg):
        "stop <started_issue> -- Stop work on an issue"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        if issue.status == PAUSED:
            error("already marked as " + STATUS[issue.status])
            return

        name = self.db.issue_name(issue)
        print "Stopping work on %s: %s" % (name, issue.title)
        comment = self.getcomment()
        if comment is None:
            return

        self.db.set_status(issue, PAUSED, comment=comment)
        print "Recorded stopping of work for", name
        return True

    def do_todo(self, arg):
        "todo [-a,all] [release] -- Generate todo list"

        self.initdb()

        flag = self.getarg(arg, 1)
        if flag in ("-a", "all"):
            release = self.getrelease(arg, 2, optional=True)
            closed = True
        else:
            release = self.getrelease(arg, 1, optional=True)
            closed = False

        print self.db.show_todo(release, closed)
        return True

    def do_unassign(self, arg):
        "unassign <assigned_issue> -- Unassign an issue from any releases"

        self.initdb()

        issue = self.getissue(arg, 1)
        if not issue:
            return

        if not issue.release:
            error("not assigned to a release")
            return

        name = self.db.issue_name(issue)
        relname = issue.release
        print "Unassigning %s: %s" % (name, issue.title)
        comment = self.getcomment()
        if comment is None:
            return

        self.db.set_release(issue, None, comment=comment)
        print "Unassigned", name, "from", relname

        return True

    def do_shell(self, arg):
        "shell -- run a system command"
        os.system(arg)
        return True

    def do_path(self, arg):
        "path -- show the root Ditz database path"

        self.initdb()
        print self.db.path
        return True

    def do_quit(self, arg):
        "quit -- quit the interactive command loop"
        sys.exit()

    do_EOF = do_quit

    # Completion methods.

    def match_issue(self, text, line, beg, end):
        return startswith(self.db.issue_names, text)

    complete_add_reference = match_issue
    complete_assign = match_issue
    complete_close = match_issue
    complete_comment = match_issue
    complete_drop = match_issue
    complete_edit = match_issue
    complete_set_component = match_issue
    complete_show = match_issue
    complete_start = match_issue
    complete_stop = match_issue
    complete_unassign = match_issue

    def match_release(self, text, line, beg, end):
        return startswith(self.db.releases, text)

    complete_archive = match_release
    complete_changelog = match_release
    complete_release = match_release
    complete_status = match_release
    complete_todo = match_release

    def match_command(self, text):
        return startswith(self.commands, text)

    # Command hook methods.

    def onecmd(self, cmd):
        args = cmd.split()

        if args:
            args[0] = args[0].replace('-', '_')

            if args[0] not in self.commands:
                matches = self.match_command(args[0])
                if len(matches) == 1:
                    args[0] = matches[0]
                elif len(matches) > 1:
                    print "Ambiguous command: %s (%s)" % \
                          (args[0], ", ".join(matches))
                    return False

        try:
            retval = Cmd.onecmd(self, " ".join(args))
        except DitzError as msg:
            error(msg)
            retval = False

        return None if self.interactive else retval

    def emptyline(self):
        return self.onecmd("todo")

    def default(self, arg):
        if self.interactive:
            print "Unknown command:", arg, "(type 'help' for a list)"
        else:
            error("unknown command: %s" % arg)

        return False

    # Input methods.

    def getarg(self, arg, idx):
        """
        Get a string argument.
        """

        args = arg.split()
        return args[idx - 1] if len(args) >= idx else None

    def getint(self, arg, idx, default=None):
        """
        Get an integer.
        """

        value = self.getarg(arg, idx)
        if not value:
            return default

        try:
            return int(value)
        except ValueError:
            error("expected an integer, not", value)
            return default

    def getissue(self, arg, idx):
        """
        Get an issue by name.
        """

        name = self.getarg(arg, idx)
        if not name and self.last_issuename:
            name = self.last_issuename

        if not name:
            error("no issue specified")
            return None

        self.last_issuename = name
        issue = self.db.get_issue(name)
        if not issue:
            error("no issue with name '%s'" % name)
            return None

        return issue

    def getrelease(self, arg, idx, optional=False):
        """
        Get a release by name.
        """

        name = self.getarg(arg, idx)
        if not name and optional:
            return None

        if not name and self.last_releasename:
            name = self.last_releasename

        if not name:
            error("no release specified")
            return None

        self.last_releasename = name
        if name not in self.db.releases:
            error("no release with name '%s'" % name)
            return None

        return name

    def getline(self, prompt="> ", default="", interrupt=True):
        """
        Get a single line of input.
        """

        if default:
            prompt = "%s (default: %s): " % (prompt, default)

        # This is a bit naughty, and copies some internals of the Cmd
        # module.  But if the function had been exposed in the first place,
        # I wouldn't have to, would I?
        if self.use_rawinput:
            try:
                return raw_input(prompt) or default
            except EOFError:
                return None
            except KeyboardInterrupt:
                if interrupt:
                    print
                    return None
                else:
                    raise
        else:
            self.stdout.write(prompt)
            self.stdout.flush()
            line = self.stdin.readline()
            if not len(line):
                return default
            else:
                return line.rstrip('\r\n')

    def gettext(self, title=None, prompt="> ", endchar='.'):
        """
        Get multiline text.
        """

        if title:
            print title + " (ctrl-c to abort, %s to finish)" % endchar

        try:
            lines = []
            while True:
                line = self.getline(prompt, interrupt=False)
                if line is None or line == endchar:
                    return "\n".join(lines)

                lines.append(line)
        except KeyboardInterrupt:
            print
            print "Aborted"
            return None

    def getchoice(self, thing, choices):
        """
        Get a choice of several things.
        """

        items = []
        for num, entry in enumerate(choices, 1):
            item = "%3d) %s" % (num, entry)
            items.append(item)

        print_columns(items)
        prompt = "Choose a %s (1--%d): " % (thing, len(choices))

        while True:
            reply = self.getline(prompt)
            if not reply:
                return None

            try:
                return choices[int(reply) - 1]
            except (ValueError, IndexError):
                pass

    def getcomment(self):
        """
        Get a comment string.
        """

        if self.nocomments:
            return ""
        elif not self.interactive and self.comment:
            return self.comment
        else:
            return self.gettext("Comments")

    def getyesno(self, question, default=False):
        """
        Get the answer to a yes/no question.
        """

        prompt = question + " [%s] " % ("yes" if default else "no")

        while True:
            reply = self.getline(prompt)
            if not reply:
                return default
            if reply[0] in "yY":
                return True
            elif reply[0] in "nN":
                return False

    def getconfig(self, oldconfig=None):
        """
        Prompt for and return database configuration info.
        """

        default = oldconfig.name if oldconfig else default_name()
        name = self.getline("Your name", default)
        if not name:
            return

        default = oldconfig.email if oldconfig else default_email()
        email = self.getline("Your email", default_email())
        if not email:
            return

        default = oldconfig.issue_dir if oldconfig else "issues"
        issuedir = self.getline("Issue directory", default)
        if not issuedir:
            return

        return Config(name, email, issuedir)

    # Miscellaneous other stuff.

    def unimplemented(self):
        cmd = self.lastcmd.split()[0]
        print "Sorry, '%s' is not implemented yet" % cmd
        return False


def startswith(items, text):
    return filter(lambda x: x.startswith(text), items)



"""
Issue, release and status flags.
"""

UNSTARTED = ":unstarted"
PAUSED = ":paused"
IN_PROGRESS = ":in_progress"
CLOSED = ":closed"
STATUS = {UNSTARTED: "unstarted",
          PAUSED: "paused",
          IN_PROGRESS: "in progress",
          CLOSED: "closed"}
FLAGS = {UNSTARTED: "_",
         PAUSED: "=",
         IN_PROGRESS: ">",
         CLOSED: "x"}
SORT = {UNSTARTED: 1,
        PAUSED: 2,
        IN_PROGRESS: 3,
        CLOSED: 0}

BUGFIX = ":bugfix"
FEATURE = ":feature"
TASK = ":task"
TYPE = {BUGFIX: "bugfix",
        FEATURE: "feature",
        TASK: "task"}
TYPE_PLURAL = {BUGFIX: "bugfixes",
               FEATURE: "features",
               TASK: "tasks"}

FIXED = ":fixed"
WONTFIX = ":wontfix"
REORG = ":reorg"
DISPOSITION = {FIXED: "fixed",
               WONTFIX: "won't fix",
               REORG: "reorganized"}

RELEASED = ":released"
UNRELEASED = ":unreleased"
RELSTATUS = {RELEASED: "released",
             UNRELEASED: "unreleased"}



"""
Logging facilities.
"""

import logging
from pkginfo import __title__

#: The logging object.
log = logging.getLogger(__title__)
log.setLevel(logging.WARNING)

# Add a stderr handler.
_handler = logging.StreamHandler()
_fmt = logging.Formatter("# %(message)s")
_handler.setFormatter(_fmt)
log.addHandler(_handler)


def set_verbose():
    log.setLevel(logging.INFO)



"""
Database exporters.
"""

import os
import re
import sys
import shutil
import codecs

from jinja2 import Environment, PackageLoader
from pkg_resources import resource_exists, resource_listdir, resource_filename

from util import make_directory
from config import config
import pkginfo
import flags


class Exporter(object):
    """
    Base class for database exporters.
    """

    #: Exporter name (for looking up templates).
    name = None

    def __init__(self, database):
        self.db = database

        # Set up Jinja templates and environment.
        templates = os.path.join('templates', self.name)
        self.loader = PackageLoader('ditz', templates)
        self.env = Environment(loader=self.loader,
                               trim_blocks=True, lstrip_blocks=True)

        # Add common filters.
        def issues(item):
            attr = item.__class__.__name__.lower()
            return [issue for issue in self.db.issues
                    if getattr(issue, attr) == item.name]

        rmap = {rel.name: rel for rel in self.db.project.releases}

        def release(issue):
            return rmap.get(issue.release, None)

        cmap = {comp.name: comp for comp in self.db.project.components}

        def component(issue):
            return cmap[issue.component]

        def dateformat(value, format="%Y-%m-%d"):
            return value.strftime(format)

        def timeformat(value, format="%Y-%m-%d %H:%M"):
            return value.strftime(format)

        self.addfilter(issues)
        self.addfilter(release)
        self.addfilter(component)
        self.addfilter(dateformat)
        self.addfilter(timeformat)

        self.setup()

    def setup(self):
        pass

    def addfilter(self, func):
        self.env.filters[func.func_name] = func

    def export(self, dirname):
        # Create output directory if required.
        make_directory(dirname)

        # Write files.
        self.write(dirname)

        # Copy static files, if any.
        static = os.path.join('static', self.name)
        if resource_exists(__name__, static):
            for name in resource_listdir(__name__, static):
                fname = os.path.join(static, name)
                src = resource_filename(__name__, fname)
                dst = os.path.join(dirname, name)
                shutil.copyfile(src, dst)

    def render(self, dirname, templatefile=None, targetfile=None,
               template=None, **kw):
        # Read template if required.
        if not template:
            template = self.env.get_template(templatefile)

        # Set target file if not specified.
        if not targetfile:
            targetfile = templatefile

        # Render template.
        text = template.render(version=pkginfo.__version__,
                               url=pkginfo.__url__,
                               **kw)

        # Write it to file.
        path = os.path.join(dirname, targetfile)
        with codecs.open(path, "w", encoding='utf-8') as fp:
            fp.write(text)

    def write(self, dirname):
        raise NotImplementedError


class HTMLExporter(Exporter):
    """
    A HTML database exporter.
    """

    name = 'html'

    def setup(self):
        def progressmeter(value, size=50):
            done = int(value * size)
            undone = max(0, size - done)
            return ("<span class='progress-meter'>" +
                    "<span class='progress-meter-done'>" +
                    ("&nbsp;" * done) +
                    "</span><span class='progress-meter-undone'>" +
                    ("&nbsp;" * undone) +
                    "</span></span>")

        def imagetag(issue, cls="inline-status-image"):
            if issue.status == flags.UNSTARTED:
                return ""

            text = flags.STATUS[issue.status]

            if issue.status == flags.CLOSED:
                text = flags.DISPOSITION[issue.disposition]

                if issue.disposition == flags.FIXED:
                    image = "green-check.png"
                elif issue.disposition == flags.WONTFIX:
                    image = "red-check.png"
                elif issue.disposition == flags.REORG:
                    image = "blue-check.png"
            elif issue.status == flags.IN_PROGRESS:
                image = "green-bar.png"
            elif issue.status == flags.PAUSED:
                image = "yellow-bar.png"

            return '<img class="%s" alt="%s" title="%s" src="%s">' % \
                   (cls, text, text, image)

        def sortkey(issue):
            text = issue.status
            if issue.closed:
                text += " " + issue.disposition
            return text

        def link(item, text=None, image=False, cls=None):
            if not item:
                return ""

            if not text:
                text = item.name

            filename = self.htmlfile(item)
            linktext = '<a href="%s">%s</a>' % (filename, text)

            if image:
                linktext += ' ' + imagetag(item)

            if cls:
                linktext = '<span class="%s">%s</span>' % (cls, linktext)

            return linktext

        idmap = {}
        for issue in self.db.issues:
            idmap[issue.id] = link(issue, image=True, cls="inline-issue-link")

        def addlinks(text):
            return self.db.convert_to_name(text, idmap)

        obscure = re.compile(r'(.+)@[\w.]+(.*)')

        def email(text):
            m = obscure.search(text)
            if m:
                return m.group(1) + '@...' + m.group(2)
            else:
                return text

        self.addfilter(link)
        self.addfilter(email)
        self.addfilter(sortkey)
        self.addfilter(addlinks)
        self.addfilter(imagetag)
        self.addfilter(progressmeter)

    def write(self, dirname):
        project = self.db.project
        showcomp = len(self.db.components) >= 2
        log = Logger()

        # Write index page.
        events = sorted(self.db.issue_events,
                        key=lambda x: x[0], reverse=True)

        log.start("Writing index page")

        count = config.getint('html', 'index_events')
        self.render(dirname, 'index.html', project=project,
                    issues=self.db.issues, activity=events[:count])

        log.finish()

        # Write a page for each release.
        log.start("Writing release pages")
        template = self.env.get_template("release.html")

        for rel in project.releases:
            path = self.htmlfile(rel)
            issues = [i for i in self.db.issues if i.release == rel.name]
            issues = sorted(issues, key=lambda x: x.creation_time)
            issues = list(reversed(issues))

            relevents = [e for e in events if e[4].release == rel.name]

            count = config.getint('html', 'release_events')
            self.render(dirname, template=template, targetfile=path,
                        project=project, release=rel, issues=issues,
                        activity=relevents[:count], show_components=showcomp)

        log.finish()

        # Write a page for each component.
        log.start("Writing component pages")
        template = self.env.get_template("component.html")

        for comp in project.components:
            path = self.htmlfile(comp)
            issues = [i for i in self.db.issues if i.component == comp.name]
            issues = sorted(issues, key=lambda x: x.creation_time)
            issues = list(reversed(issues))

            self.render(dirname, template=template, targetfile=path,
                        project=project, component=comp, issues=issues)

        log.finish()

        # Write a page for each issue.
        log.start("Writing issue pages")
        total = len(self.db.issues)
        template = self.env.get_template("issue.html")

        for num, issue in enumerate(self.db.issues, 1):
            log.percent(100.0 * num / total)
            path = self.htmlfile(issue)

            self.render(dirname, template=template, targetfile=path,
                        project=project, issue=issue)

        log.finish()

        # Write a page for unassigned issues.
        log.start("Writing unassigned issue page")
        issues = [i for i in self.db.issues if not i.release]
        issues = sorted(issues, key=lambda x: x.creation_time)
        issues = list(reversed(issues))

        self.render(dirname, 'unassigned.html', project=project,
                    issues=issues, show_components=showcomp)

        log.finish()

        # Write index file location.
        path = os.path.join(dirname, 'index.html')
        log.status("Local generated URL: file://%s" % os.path.abspath(path))

    def htmlfile(self, item):
        clsname = item.__class__.__name__
        name = getattr(item, "id", None) or getattr(item, "name")
        return "%s-%s.html" % (clsname.lower(), name)


class Logger(object):
    def __init__(self, stream=sys.stdout, dots="... "):
        self.stream = stream
        self.dots = dots

    def status(self, msg, newline=True):
        self.stream.write(msg + ("\n" if newline else ""))
        self.stream.flush()

    def start(self, msg):
        self.msg = msg

    def percent(self, value):
        self.status(self.msg + self.dots + ("%.0f%%" % value) + '\r', False)

    def finish(self):
        self.status(self.msg + self.dots + "done")



"""
Configuration file.
"""

import os
from ConfigParser import RawConfigParser

from pkginfo import __title__
from logger import log


class DitzConfig(RawConfigParser):
    def __init__(self):
        RawConfigParser.__init__(self)

        self.add_defaults("install",
                          command="pyditz")

        self.add_defaults("html",
                          index_events=10,
                          release_events=10)

    def add_defaults(self, section, **defaults):
        """Add a section with defaults."""

        self.add_section(section)
        for key, val in defaults.items():
            self.set(section, key, val)

    def write_file(self, path):
        """Write config data to file."""

        with open(path, "w") as fp:
            fp.write("# %s configuration file.\n\n" % __title__)
            self.write(fp)


class Config(object):
    def __init__(self):
        homedir = os.path.expanduser("~")
        self.path = os.path.join(homedir, ".ditzrc")
        self.parser = None

    def set_file(self, path):
        self.path = path

    def __getattr__(self, attr):
        if not self.parser:
            self.parser = DitzConfig()
            log.info("reading %s" % self.path)
            self.parser.read(self.path)

        return getattr(self.parser, attr)


# Ditz user config settings.
config = Config()



"""
Database tests.
"""

import shutil
from filecmp import dircmp
from nose.tools import raises

from ditz.database import DitzDB, DitzError
from ditz.flags import *

desc = """
blah blah blah blah blah blah blah blah blah blah blah blah blah
blah blah blah blah blah blah blah blah blah blah blah blah blah
blah blah blah blah blah blah blah blah blah blah blah blah blah
"""


def test_create(name="database"):
    "Test database creation"

    try:
        shutil.rmtree(name)
    except OSError:
        pass

    db = DitzDB("MyProject", path=name, usecache=True, autosave=True)
    print db

    db.add_component("default")
    db.add_release('0.1')
    db.add_release('0.2')

    issue = db.add_issue("Something to do", desc)
    db.add_reference(issue, "file.txt")

    issue = db.add_issue("Critical fix", desc, comment="Really urgent")

    db.set_release(issue, '0.1')
    db.set_release(issue, '0.2')
    db.set_release(issue, None)

    db.add_comment(issue, "Will you make up your mind?")

    db.set_status(issue, IN_PROGRESS, comment="Let's do it!")
    db.set_status(issue, PAUSED, comment="Er.. let's think about it")
    db.set_status(issue, CLOSED, WONTFIX, comment="Nah, let's not bother")

    for issue in db:
        print issue

    db.write()

    issue = db.add_issue("Another issue")
    db.write()

    db.drop_issue(issue)
    db.write()

    newdb = DitzDB.read(name, usecache=True)
    assert len(newdb.issues) == 2

    newdb = DitzDB.read(name, usecache=False)
    assert len(newdb.issues) == 2


@raises(DitzError)
def test_nosuchcomp():
    "Test nonexistent component"

    db = DitzDB("MyProject")
    db.add_issue("Issue 1", component='fred')


def test_release():
    "Test valid release"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.release_release("0.1")


@raises(DitzError)
def test_norelease():
    "Test nonexistent release"

    db = DitzDB("MyProject")
    db.add_component("default")
    db.add_issue("Issue 1", release='1.0')


@raises(DitzError)
def test_duprelease():
    "Test duplicate releases"

    db = DitzDB("MyProject")
    db.add_release("0.2")
    db.add_release("0.2")


@raises(DitzError)
def test_invalidrelease_1():
    "Test invalid release (no issues)"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.release_release("0.1")


@raises(DitzError)
def test_invalidrelease_2():
    "Test invalid release (issue not closed)"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1')
    db.release_release("0.1")


@raises(DitzError)
def test_invalidrelease_3():
    "Test invalid release (no such release)"

    db = DitzDB("MyProject")
    db.release_release("0.1")


@raises(DitzError)
def test_invalidrelease_4():
    "Test invalid release (already released)"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.release_release("0.1")
    db.release_release("0.1")


@raises(DitzError)
def test_dupcomponent():
    "Test duplicate components"

    db = DitzDB("MyProject")
    db.add_component("doc")
    db.add_component("doc")


def test_show_todo():
    "Test show todo"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.show_todo()
    db.show_todo("0.1")
    db.show_todo(None, True)


def test_show_changelog():
    "Test show changelog"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.add_issue("Issue 2", release='0.1', status=CLOSED, type=FEATURE)
    db.show_changelog("0.1")
    db.release_release("0.1")
    db.show_changelog("0.1")


def test_archive():
    "Test show changelog"

    db = DitzDB("MyProject")
    db.add_release("0.1")
    db.add_issue("Issue 1", release='0.1', status=CLOSED)
    db.release_release("0.1")
    db.archive_release("0.1", "example-archive")


def test_copy(name="example"):
    "Test database copying and roundtripping"

    db = DitzDB.read(name)
    db.write(name + "-copy")

    def check_files(dcmp):
        for name in dcmp.diff_files:
            raise Exception("file '%s' differs" % name)

        for sub_dcmp in dcmp.subdirs.values():
            check_files(sub_dcmp)

    dcmp = dircmp(name, name + "-copy")
    check_files(dcmp)


def test_convert():
    "Test conversion between names and IDs in strings"

    db = DitzDB("MyProject")
    issue = db.add_issue("A bug")

    text = db.convert_to_id("This is issue ID myproject-1")
    assert text == "This is issue ID {issue %s}" % issue.id

    text = db.convert_to_name("This is issue name {issue %s}" % issue.id)
    assert text == "This is issue name myproject-1"

    name = db.issue_name(issue)
    assert name == "myproject-1"

    otherissue = db.get_issue(name)
    assert otherissue == issue



"""
Package information.
"""

import version

__title__ = "pyditz"
__author__ = 'Glenn Hutchings'
__email__ = 'zondo42@gmail.com'
__version__ = version.__doc__.strip()
__license__ = 'GPL v2 or later'
__copyright__ = "2013,2014, " + __author__
__url__ = "http://pypi.python.org/pypi/pyditz"

__classifiers__ = """
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)
Natural Language :: English
Operating System :: Microsoft :: Windows
Operating System :: POSIX :: Linux
Operating System :: Unix
Programming Language :: Python :: 2.7
Topic :: Software Development :: Bug Tracking
"""



"""
Test utility functions.
"""

from ditz import util


def test_utils():
    "Test utility functions"

    for num in 0.43, 1, 23, 64, 432, 3656, 1539424, 41749360, 97492847:
        print num, '=', util.ago(num)

    print
    print util.extract_username('Alan Partridge <alan@norwichradio.co.uk>')

    print
    names = "fred brian derek tarquin ermintrude cyril"
    util.print_columns(names.split(), linelen=45)

    print
    print util.default_name()
    print util.default_email()


if __name__ == "__main__":
    test_utils()



"""
Ditz database interface.
"""

import os
import re
import glob
import cPickle as pickle

from collections import Counter
from datetime import datetime

from objects import Config, Project, Issue, Component, Release, find_config

from util import (read_yaml_file, age, extract_username, default_name,
                  default_email, DitzError)

from exporter import HTMLExporter
from logger import log

from flags import (RELEASED, UNRELEASED, BUGFIX, FEATURE, TASK, UNSTARTED,
                   STATUS, DISPOSITION, TYPE, TYPE_PLURAL, CLOSED, FLAGS)


class DitzDB(object):
    def __init__(self, project, username=None, email=None,
                 issuedir=None, path=".", autosave=False,
                 usecache=False, cachefile=".ditz-cache"):
        #: Configuration data.
        self.config = Config(username or default_name(),
                             email or default_email(),
                             issuedir or ".ditz-issues")

        #: Location.
        self.path = os.path.realpath(path)

        #: Project data.
        self.project = Project(project)

        #: List of issues.
        self.issues = []

        #: Whether to save files when changes are made.
        self.autosave = autosave

        #: Whether to use issue cache.
        self.usecache = usecache

        #: Issue cache filename.
        self.cachefile = cachefile

        #: Mapping of issue IDs to assigned names.
        self._id2name = {}

        #: Mapping of assigned names to issues.
        self._name2issue = {}

    # High-level database commands.

    def add_issue(self, title, desc="", type=BUGFIX, status=UNSTARTED,
                  disposition=None, creation_time=None, reporter=None,
                  component=None, release=None, comment=None):
        """
        Add a new issue.
        """

        issue = Issue(title, desc, type, status, disposition, creation_time,
                      reporter or self.config.username)

        self.set_component(issue, component, event=False)
        self.set_release(issue, release, event=False)

        self.issues.append(issue)
        self.add_event(issue, "created", comment)
        self.update_issue_names()

        self.save(issue)
        return issue

    def add_component(self, name):
        """
        Add a new component.
        """

        comp = Component(name)

        if name not in self.components:
            self.project.components.append(comp)
        else:
            raise DitzError("component already exists: %s" % name)

        self.save()
        return comp

    def add_release(self, name, status=UNRELEASED, release_time=None,
                    comment=None):
        """
        Add a new release.
        """

        if not self.get_release(name):
            rel = Release(name, status, release_time)
            self.add_event(rel, "created", comment)
            self.project.releases.append(rel)
        else:
            raise DitzError("release already exists: %s" % name)

        self.save()
        return rel

    def add_reference(self, issue, reference, comment=None):
        """
        Add a reference to an issue.
        """

        num = issue.add_reference(reference)
        self.add_event(issue, "added reference %d" % num, comment)
        self.save(issue)

    def set_status(self, issue, status, disposition=None, comment=None):
        """
        Set the status of an issue.
        """

        prevstatus = issue.status

        issue.set_status(status)
        issue.set_disposition(disposition)

        if disposition:
            text = "closed with disposition %s" % DISPOSITION[disposition]
        else:
            text = "changed status from %s to %s" % \
                   (STATUS[prevstatus], STATUS[status])

        self.add_event(issue, text, comment)
        self.save(issue)

    def set_component(self, issue, component=None, comment=None, event=True):
        """
        Set the component of an issue.
        """

        if not component:
            component = self.project.name
        elif component not in self.components:
            raise DitzError("unknown component: %s" % component)

        if event:
            text = "assigned to component %s from %s" \
                   % (component, issue.component)
            self.add_event(issue, text, comment)

        issue.component = component
        self.update_issue_names()
        self.save(issue)

    def set_release(self, issue, release=None, comment=None, event=True):
        """
        Set the release of an issue.
        """

        if release and not self.get_release(release):
            raise DitzError("unknown release: %s" % release)

        if issue.release == release:
            return

        if event:
            if release:
                text = "assigned to release " + release
                if issue.release:
                    text += " from release " + issue.release
                else:
                    text += " from unassigned"
            else:
                text = "unassigned from release %s" % issue.release

            self.add_event(issue, text, comment)

        issue.release = release
        self.save(issue)

    def release_release(self, name, comment=None):
        """
        Release a release.
        """

        rel = self.get_release(name)
        if not rel:
            raise DitzError("unknown release: %s" % name)

        if rel.status == RELEASED:
            raise DitzError("release '%s' is already released" % name)

        count = 0
        for issue in self.issues:
            if issue.release == name:
                count += 1
                if issue.status != CLOSED:
                    raise DitzError("open issue %s must be reassigned"
                                    % self.issue_name(issue))

        if count == 0:
            raise DitzError("no issues assigned to release '%s'" % name)

        rel.status = RELEASED
        rel.release_time = datetime.now()

        self.add_event(rel, "released", comment)
        self.save()

    def archive_release(self, name, path):
        """
        Archive a release.
        """

        rel = self.get_release(name)
        if not rel:
            raise DitzError("unknown release: %s" % name)

        if rel.status != RELEASED:
            raise DitzError("release '%s' has not been released" % name)

        keep = []
        archive = []
        for issue in self.issues:
            if issue.release == name:
                archive.append(issue)
            else:
                keep.append(issue)

        if not archive:
            raise DitzError("no issues assigned to release %s" % name)

        try:
            os.makedirs(path)
        except OSError as msg:
            raise DitzError("can't create %s: %s" % (path, str(msg)))

        self.project.write(path)
        for issue in archive:
            issue.write(path)

        for issue in archive:
            path = self.issue_filename(issue)
            if os.path.exists(path):
                os.remove(path)

        self.issues = keep
        self.update_issue_names()

        self.save()

    def drop_issue(self, issue):
        """
        Drop an issue.
        """

        path = self.issue_filename(issue)
        if os.path.exists(path):
            os.remove(path)

        self.issues.remove(issue)
        self.update_issue_names()

        self.save()

    def add_comment(self, issue, comment):
        """
        Add a comment to an issue.
        """

        self.add_event(issue, "commented", comment)
        self.save(issue)

    def show_grep(self, regexp):
        """
        Return text describing issues that match a regexp.
        """

        issues = [issue for issue in self.issues if issue.grep(regexp)]

        if issues:
            return "\n".join(self.list_issues(issues))
        else:
            return None

    def show_todo(self, release=None, closed=False):
        """
        Return text describing unresolved issues.
        """

        if not release:
            releases = [r for r in self.project.releases
                        if r.status == UNRELEASED] + [None]
        else:
            releases = [r for r in self.project.releases
                        if r.name == release]

        lines = []
        add = lines.append

        for num, rel in enumerate(releases):
            if num > 0:
                add('')

            if rel:
                text = rel.name
                if rel.status == UNRELEASED:
                    text += " (unreleased)"
                add(text + ":")
            else:
                add("Unassigned:")

            show = []
            for issue in self.issues:
                if rel and issue.release != rel.name:
                    continue
                elif not rel and issue.release:
                    continue
                elif not closed and issue.status == CLOSED:
                    continue

                show.append(issue)

            if not show:
                add("No issues" if closed else "No open issues")
                continue

            for text in self.list_issues(sorted(show, reverse=True)):
                add(text)

        return "\n".join(lines)

    def show_issue(self, issue, attrfmt="%11s: %s", underline='-'):
        """
        Return text describing an issue.
        """

        lines = []
        add = lines.append
        addattr = lambda attr, val="": add(attrfmt % (attr, val))

        title = "Issue %s" % self.issue_name(issue)
        add(title)
        add(underline * len(title))

        addattr("Title", issue.title)

        desc = self.convert_to_name(issue.desc)

        if "\n" in desc:
            addattr("Description")
            for line in desc.split("\n"):
                add("  " + line)
        else:
            addattr("Description", desc)

        add('')
        addattr("Type", TYPE[issue.type])

        if issue.status == CLOSED:
            addattr("Status", "closed: " + DISPOSITION[issue.disposition])
        else:
            addattr("Status", STATUS[issue.status])

        addattr("Creator", issue.reporter)
        addattr("Age", age(issue.creation_time))
        addattr("Release", issue.release or "")

        addattr("References")
        for num, ref in enumerate(issue.references, 1):
            add("%3d. %s" % (num, ref))

        addattr("Identifier", issue.id)

        add('')
        add("Event log:")

        for date, email, text, comment in reversed(issue.log_events):
            add("- %s (%s, %s ago)" % (text, extract_username(email),
                                       age(date)))
            if comment:
                for line in self.convert_to_name(comment).split("\n"):
                    add('  > ' + line)

        return "\n".join(lines)

    def show_releases(self):
        """
        Return text describing all releases.
        """

        lines = []
        add = lines.append

        for rel in reversed(self.project.releases):
            if rel.status == UNRELEASED:
                tag = "unreleased"
            else:
                tag = "released %s" % (rel.release_time.strftime("%Y-%m-%d"))

            add("%s (%s)" % (rel.name, tag))

        return "\n".join(lines)

    def show_changelog(self, name):
        """
        Return text describing a changelog for a release.
        """

        rel = self.get_release(name)
        if not rel:
            raise DitzError("unknown release: %s" % name)

        lines = []
        add = lines.append

        text = "== " + rel.name + " / "
        if rel.status == RELEASED:
            text += rel.release_time.strftime("%Y-%m-%d")
        else:
            text += "unreleased"

        add(text)

        for issue in sorted(self.issues, key=lambda x: x.type):
            if issue.release == rel.name:
                if issue.status != CLOSED:
                    continue

                if issue.type == BUGFIX:
                    text = "bugfix: " + issue.title
                else:
                    text = issue.title

                add("* " + text)

        return "\n".join(lines)

    def show_status(self, release=None, maxflags=20):
        """
        Return text describing the status of a release or releases.
        """

        if not release:
            releases = [r for r in self.project.releases
                        if r.status == UNRELEASED] + [None]
        else:
            releases = [r for r in self.project.releases
                        if r.name == release]

        data = []
        for rel in releases:
            reldata = []
            data.append(reldata)
            reldata.append(rel.name if rel else "unassigned")

            issues = []
            alltotal = allclosed = 0
            for itype in BUGFIX, FEATURE, TASK:
                closed = total = 0

                for issue in self.issues:
                    if issue.type != itype:
                        continue
                    elif rel and issue.release != rel.name:
                        continue
                    elif not rel and issue.release:
                        continue

                    total += 1
                    if issue.status == CLOSED:
                        closed += 1

                    issues.append(issue)

                alltotal += total
                allclosed += closed

                text = "%2d/%2d %s" % (closed, total, TYPE_PLURAL[itype])
                reldata.append(text)

            if not rel:
                text = ""
            elif rel.status == RELEASED:
                text = "(released)"
            elif alltotal == 0:
                text = "(no issues)"
            elif allclosed == alltotal:
                text = "(ready for release)"
            else:
                flags = map(lambda x: FLAGS[x.status], sorted(issues))
                if len(flags) > maxflags:
                    newflags = []

                    for i in xrange(maxflags):
                        factor = float(i) / (maxflags - 1)
                        idx = int(factor * (len(flags) - 1))
                        newflags.append(flags[idx])

                    flags = newflags

                text = "".join(flags)

            reldata.append(text)

        maxlen = [0] * 5
        for reldata in data:
            for col in xrange(5):
                maxlen[col] = max(maxlen[col], len(reldata[col]))

        lines = []
        for reldata in data:
            for col in xrange(5):
                if col in (0, 4):
                    reldata[col] = reldata[col].ljust(maxlen[col])
                else:
                    reldata[col] = reldata[col].rjust(maxlen[col])

            lines.append("  ".join(reldata))

        return "\n".join(lines)

    def export_html(self, path):
        """
        Export HTML summary to the specified directory.
        """

        exporter = HTMLExporter(self)
        exporter.export(path)

    def log_events(self, verbose=False, count=None, datefmt="%a %b %d %X %Y"):
        """
        Yield log event messages.
        """

        events = sorted(self.issue_events, key=lambda x: x[0], reverse=True)
        for num, (date, email, text, comment, issue) in enumerate(events):
            if count and num >= count:
                break

            name = self.issue_name(issue)
            when = age(date)

            if verbose:
                lines = []
                add = lines.append

                add('date   : %s (%s ago)' % (date.strftime(datefmt), when))
                add('author : %s' % email)
                add('issue  : [%s] %s' % (name, issue.title))
                add('')
                add('  ' + text)

                if comment:
                    for line in self.convert_to_name(comment).split("\n"):
                        add('  > ' + line)

                add('')

                yield "\n".join(lines)
            else:
                user = extract_username(email)
                yield "%10s | %10s | %15s | %s" % (when, name, user, text)

    # Low-level commands and properties.

    def add_event(self, item, text, comment=None):
        """
        Add an event to a database item.
        """

        if comment:
            comment = self.convert_to_id(comment)

        item.event(self.config.username, text, comment)

    def get_issue(self, name):
        """
        Return an issue given its assigned issue name.
        """

        return self._name2issue.get(name, None)

    def issue_name(self, issue):
        """
        Return the assigned issue name of an issue.
        """

        return self._id2name.get(issue.id, "<invalid-issue>")

    def issue_filename(self, issue):
        """
        Return the filename of an issue.
        """

        return os.path.join(self.path, self.config.issue_dir, issue.filename)

    def get_release(self, name):
        """
        Return a release given its name.
        """

        name = str(name)
        for rel in self.project.releases:
            if rel.name == name:
                return rel

        return None

    def list_issues(self, issues):
        """
        Return text description lines for a list of issues.
        """

        lines = []
        maxlen = max(len(self.issue_name(issue)) for issue in issues)

        for issue in issues:
            text = FLAGS[issue.status] + " "
            text += self.issue_name(issue).rjust(maxlen)
            text += ": " + issue.name
            lines.append(text)

        return lines

    @property
    def issue_names(self):
        """
        All existing issue names.
        """

        return sorted(self._name2issue.keys())

    @property
    def issue_events(self):
        """
        Yield all issue-related events and their issues.
        """

        for issue in self.issues:
            for date, user, text, comment in issue.log_events:
                yield date, user, text, comment, issue

    @property
    def components(self):
        """
        List of defined components.
        """

        return [comp.name for comp in self.project.components]

    @property
    def releases(self):
        """
        List of defined releases.
        """

        return [rel.name for rel in self.project.releases]

    # Database input/output methods.

    @staticmethod
    def read(dirname, **kwargs):
        """
        Read a Ditz database.
        """

        # Create an empty database.
        db = DitzDB("empty", **kwargs)

        # Find the Ditz config file.
        path = find_config(dirname, error=True)
        dirname = os.path.split(path)[0]

        # Read it.
        log.info("reading config from %s" % path)
        config = read_yaml_file(path)
        config.validate()

        # Extract the issue directory name.
        try:
            issuedir = config.issue_dir
        except AttributeError:
            raise DitzError("'%s' does not define 'issue_dir'" % path)

        # Read the project file.
        path = os.path.join(dirname, issuedir, Project.filename)
        log.info("reading project from %s" % path)
        project = read_yaml_file(path)
        project.validate()

        # Read the issues.
        issues = []
        match = os.path.join(dirname, issuedir, Issue.template % "*")
        issue_files = set(glob.glob(match))

        path = os.path.join(dirname, db.cachefile)
        cached_issues = db.readcache(path)
        cache_changed = False

        if cached_issues:
            # Check for new, modified or deleted issues.
            cached_files = set()
            modtime = os.stat(path).st_mtime

            for issue in cached_issues:
                path = os.path.join(dirname, issuedir, issue.filename)

                # Skip issue if it's deleted.
                if not os.path.exists(path):
                    continue

                # Reread issue file if invalid, or newer than cache.
                reread = os.stat(path).st_mtime > modtime

                try:
                    issue.validate()
                except DitzError:
                    reread = True

                if reread:
                    log.info("reading changed issue from %s" % path)
                    issue = read_yaml_file(path)
                    issue.validate()
                    cache_changed = True

                cached_files.add(path)
                issues.append(issue)

            # Read all the new issues.
            for path in issue_files - cached_files:
                log.info("reading new issue from %s" % path)
                issue = read_yaml_file(path)
                issue.validate()
                issues.append(issue)
                cache_changed = True
        else:
            log.info("reading issues from %s" % issuedir)
            for path in issue_files:
                issue = read_yaml_file(path)
                issue.validate()
                issues.append(issue)
                cache_changed = True

        # Initialise database.
        db.path = os.path.realpath(dirname)
        db.config = config
        db.project = project
        db.issues = issues

        if cache_changed:
            db.update_cache()

        db.update_issue_names()

        return db

    def readcache(self, path):
        """
        Read cached issues if possible.
        """

        if not self.usecache or not os.path.exists(path):
            return None

        log.info("reading issue cache from %s" % path)

        with open(path) as fp:
            try:
                issues = pickle.load(fp)
            except Exception as msg:
                log.info("removing issue cache (%s)" % str(msg))
                os.unlink(path)
                issues = None

        return issues

    def write(self, dirname=None):
        """
        Write the database.
        """

        if not dirname:
            dirname = self.path
        else:
            self.path = dirname

        # Write the config file.
        self.config.write(dirname)

        # Write the project file.
        path = os.path.join(dirname, self.config.issue_dir)
        self.project.write(path)

        # Write the issues.
        for issue in self.issues:
            issue.write(path)

        # Update issue cache.
        self.update_cache()

    def save(self, issue=None):
        """
        Save database file if required.
        """

        if not self.autosave:
            return

        if issue:
            log.info("saving issue %s" % self.issue_name(issue))
            issue.write(self.issuedir)
            self.update_cache()
        else:
            log.info("saving project")
            self.project.write(self.issuedir)

    def update_cache(self):
        """
        Update the issue cache.
        """

        if self.usecache:
            path = os.path.join(self.path, self.cachefile)
            log.info("writing issue cache to %s" % path)
            with open(path, "w") as fp:
                pickle.dump(self.issues, fp)

    # Issue name/ID handling methods.

    def convert_to_id(self, text):
        """
        Replace names with {issue ...} in text.
        """

        def repl(m):
            issue = self._name2issue.get(m.group(1), None)

            if issue:
                return "{issue %s}" % issue.id
            else:
                return m.group(0)

        return re.sub(r'\b([A-Za-z]+-[0-9]+)\b', repl, text)

    def convert_to_name(self, text, idmap={}):
        """
        Replace {issue ...} with names in text.
        """

        def repl(m):
            idx = m.group(1)
            if idx in idmap:
                return idmap[idx]
            else:
                return self._id2name[idx]

        return re.sub(r'{issue ([0-9a-f]+)}', repl, text)

    def update_issue_names(self):
        """
        Reassign issue names after changes to issue status.
        """

        log.info("updating issue names")
        self.issues.sort()
        counts = Counter()

        self._id2name = {}
        self._name2issue = {}

        for issue in sorted(self.issues, key=lambda x: x.creation_time):
            comp = issue.component.lower()
            counts[comp] += 1
            name = "%s-%d" % (comp, counts[comp])
            self._id2name[issue.id] = name
            self._name2issue[name] = issue

    # Miscellaneous other stuff.

    @property
    def issuedir(self):
        return os.path.join(self.path, self.config.issue_dir)

    def __iter__(self):
        return iter(self.issues)

    def __repr__(self):
        return "<DitzDB: %s>" % self.path



"""
Main console program.
"""

import sys
import argparse

from commands import DitzCmd
from config import config
from logger import set_verbose
from pkginfo import __version__


def main(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(prog="pyditz")

    parser.add_argument("--version", action="version",
                        version="%(prog)s version " + __version__)

    group = parser.add_argument_group("global arguments")

    group.add_argument("-u", "--user-config", type=str, metavar="FILE",
                       help="use the given user config file")

    group.add_argument("-c", "--config-file", type=str, metavar="FILE",
                       help="use the given Ditz config file [NOT IMPLEMENTED]")

    group.add_argument("-i", "--issue-dir", type=str, metavar="DIR",
                       help="use the given issue directory [NOT IMPLEMENTED]")

    group.add_argument("-v", "--verbose", action="store_true",
                       help="be verbose about things")

    group = parser.add_argument_group("command arguments")

    group.add_argument("-m", "--comment", type=str, metavar="STRING",
                       help="specify a comment")

    group.add_argument("-n", "--no-comment", action="store_true",
                       help="skip asking for a comment")

    opts, args = parser.parse_known_args(args)
    cmdopts = dict(usecache=True, autosave=True, nocomment=opts.no_comment,
                   comment=opts.comment)

    if opts.user_config:
        config.set_file(opts.user_config)

    if opts.verbose:
        set_verbose()

    if not args:
        cmd = DitzCmd(interactive=True, **cmdopts)

        try:
            cmd.cmdloop()
        except KeyboardInterrupt:
            pass
    else:
        cmd = DitzCmd(**cmdopts)
        command = " ".join(args)
        if not cmd.onecmd(command):
            sys.exit(99)

if __name__ == "__main__":
    main()



"""
Command tests.
"""

from ditz.database import DitzDB
from ditz.commands import DitzCmd
from ditz.util import read_yaml_file
from mock import Mock


def test_commands(name="example"):
    "Test commands"

    dbname = name + "-cmd"
    db = DitzDB.read(name)
    db.write(dbname)

    for test in read_yaml_file("test-cmd.yaml"):
        cmd = test.pop("cmd")
        for testcase in test.pop("tests"):
            args = testcase.pop("args") or ""
            command = cmd + " " + str(args)
            print "=== Running '%s' ===" % command
            run_command(dbname, command, **testcase)


def run_command(dbname, cmd, **kw):
    ditz = DitzCmd(dbname)

    for method in "line", "text", "choice", "yesno":
        if method in kw:
            effect = kw[method]
        else:
            effect = Exception('no %s input' % method)

        setattr(ditz, "get" + method, Mock(spec=DitzCmd, side_effect=effect))

    ditz.onecmd(cmd)


if __name__ == "__main__":
    test_commands()



# -*- coding: utf-8 -*-

"""
Utility functions.
"""

import os
import re
import yaml
import socket
from datetime import datetime


def read_yaml_file(path):
    """
    Read YAML data from a file.
    """

    with open(path) as fp:
        return read_yaml(fp)


def read_yaml(fp):
    """
    Read YAML data from a stream.
    """

    return yaml.safe_load(fp)


def age(date):
    """
    Return a human-readable in-the-past time string given a date.
    """

    return ago((datetime.now() - date).total_seconds())


def ago(counter):
    """
    Return how long ago a number of seconds is.
    """

    counter = int(counter)
    if counter == 0:
        return "now"

    for unit, count in (("second", 60),
                        ("minute", 60),
                        ("hour",   24),
                        ("day",     7),
                        ("week",    4),
                        ("month",  12),
                        ("year",    0)):
        if count > 0 and counter >= count * 2:
            counter /= count
        else:
            break

    return "%d %s%s" % (counter, unit, "s" if counter > 1 else "")


def extract_username(email):
    """
    Return a short user name given email text.
    """

    if '@' not in email:
        return email

    m = re.search('([A-Za-z0-9_.]+)@', email)
    if m:
        return m.group(1)

    return email


def default_name():
    """
    Return a default user name.
    """

    for var in "DITZUSER", "USER", "USERNAME":
        if var in os.environ:
            return os.environ[var]

    return "ditzuser"


def default_email():
    """
    Return a default email address.
    """

    return default_name() + '@' + hostname()


def hostname():
    """
    Return the host name.
    """

    for var in "DITZHOST", "HOSTNAME", "COMPUTERNAME":
        if var in os.environ:
            return os.environ[var]

    try:
        return socket.gethostname()
    except socket.error:
        return "UNKNOWN"


def print_columns(items, linelen=70, spacing=2):
    """
    Print a number of items in column format.
    """

    maxlen = max(len(text) for text in items)
    columns = max(linelen / (maxlen + spacing), 1)
    padding = " " * spacing

    count = 0
    while count < len(items):
        print items[count].ljust(maxlen) + padding,
        count += 1
        if count % columns == 0:
            print

    if count % columns != 0:
        print


def make_directory(path):
    """
    Create a directory if it doesn't exist.
    """

    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as msg:
            raise DitzError(msg)


def check_value(name, value, choices):
    if value not in choices:
        raise ValueError("unknown %s: %s (one of %s expected)"
                         % (name, value, ", ".join(choices.keys())))


def error(*args):
    """
    Give a Ditz error.
    """

    print "Error: " + " ".join(map(str, args))


class DitzError(Exception):
    "A generic Ditz error."



# Setup script for PyDitz.

import ez_setup
ez_setup.use_setuptools()

from setuptools import setup

import ditz
import ditz.pkginfo as info
from ditz.config import config

try:
    from hgtools.managers import MercurialManager
    version = MercurialManager().get_current_version()
    with open("ditz/version.py", "w") as fp:
        fp.write("'%s'\n" % version)
except ImportError:
    pass


setup(name             = info.__title__,
      author           = info.__author__,
      author_email     = info.__email__,
      description      = ditz.__doc__.strip(),
      long_description = open("README").read(),
      license          = info.__license__,
      url              = info.__url__,
      classifiers      = info.__classifiers__.strip().split("\n"),

      packages = ["ditz"],
      include_package_data = True,
      use_vcs_version = True,

      setup_requires = [
          'hgtools',
      ],

      install_requires = [
          'pyyaml >= 3.10',
          'jinja2 >= 2.7',
      ],

      tests_require = [
          'nose >= 1.3.0',
          'coverage >= 3.6',
          'mock >= 1.0.1',
      ],

      test_suite = 'nose.collector',

      entry_points = {
          'console_scripts': [
              '%s = ditz.console:main' % config.get('install', 'command')
          ]
      })



"""
Main console test.
"""

import os
from ditz.console import main


def test_run():
    "Test main application"

    os.chdir("example")
    main(["todo"])
    os.chdir("..")


if __name__ == "__main__":
    test_run()



"""
Internal Ditz objects.
"""

import os
import sys
import yaml
import hashlib
import random
import codecs
from datetime import datetime

from util import make_directory, check_value, DitzError
from flags import (STATUS, RELSTATUS, RELEASED, UNRELEASED, BUGFIX,
                   FEATURE, DISPOSITION, UNSTARTED, TYPE, SORT, CLOSED)


class DitzObject(yaml.YAMLObject):
    """
    Base class of a Ditz object appearing in a YAML file.

    Attributes:
        ditz_tag (str): Base YAML tag prefix.
        yaml_tag (str): YAML tag written to file.
        filename (str): YAML filename to read/write.
        attributes (list): List of recognized attributes.
        log_events (list): List of events.
    """

    yaml_loader = yaml.SafeLoader

    ditz_tag = "!ditz.rubyforge.org,2008-03-06"
    yaml_tag = None
    filename = None
    attributes = []

    def write(self, dirname="."):
        """
        Write object to its YAML file.

        Args:
            dirname (str, optional): Directory to write file into.
        """

        make_directory(dirname)
        path = os.path.join(dirname, self.filename)

        with codecs.open(path, "wb", encoding='utf-8') as fp:
            fp.write("--- ")
            write_yaml(self, fp)

    def event(self, username, text, comment=None, timestamp=None):
        """
        Add an event to the object's ``log_events`` list.

        Args:
            username (str): User name.
            text (str): Event description.
            comment (str, optional): User comment.
            timestamp (datetime, optional): Time of the event.
        """

        time = timestamp or datetime.now()
        self.log_events.append([time, username, text, comment or ""])

    def validate(self):
        """
        Check the object is a valid Ditz object.
        """

        events = getattr(self, "log_events", [])
        for num, data in enumerate(events, 1):
            if len(data) != 4:
                self._invalid("log_events entry %d has size %d instead of 4"
                              % (num, len(data)))

    def _invalid(self, msg):
        raise DitzError("%s: %s" % (self.filename, str(msg)))


class Project(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/project'
    filename = "project.yaml"
    attributes = ["name", "version", "components", "releases"]

    def __init__(self, name, version=0.5):
        self.name = name
        self.version = version
        self.components = [Component(name)]
        self.releases = []


class Release(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/release'
    attributes = ["name", "status", "release_time", "log_events"]

    def __init__(self, name, status=UNRELEASED, release_time=None):
        self.name = name
        self.status = status
        self.release_time = release_time
        self.log_events = []

        self.set_status(status)

    def set_status(self, status):
        check_value("status", status, RELSTATUS)
        self.status = status

    @property
    def released(self):
        return self.status == RELEASED

    def __repr__(self):
        return "<Release: %s>" % self.name


class Component(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/component'
    attributes = ["name"]

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "<Component: %s>" % self.name


class Issue(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/issue'
    template = "issue-%s.yaml"
    attributes = ["title", "desc", "type", "component", "release",
                  "reporter", "status", "disposition", "creation_time",
                  "references", "id", "log_events"]

    def __init__(self, title, desc="", type=BUGFIX, status=UNSTARTED,
                 disposition=None, creation_time=None, reporter=""):
        self.title = title.strip()
        self.desc = desc.strip()
        self.component = None
        self.release = None
        self.reporter = reporter

        self.set_type(type)
        self.set_status(status)
        self.set_disposition(disposition)

        self.creation_time = creation_time or datetime.now()
        self.id = self.make_id()

        self.references = []
        self.log_events = []

    @property
    def name(self):
        if self.type == BUGFIX:
            return self.title + " (bug)"
        elif self.type == FEATURE:
            return self.title + " (feature)"
        else:
            return self.title

    @property
    def closed(self):
        return self.status == CLOSED

    @property
    def filename(self):
        return self.template % self.id

    def set_type(self, type):
        check_value("type", type, TYPE)
        self.type = type

    def set_status(self, status):
        check_value("status", status, STATUS)
        self.status = status

    def set_disposition(self, disposition):
        if disposition:
            check_value("disposition", disposition, DISPOSITION)

        self.disposition = disposition

    def add_reference(self, reference):
        self.references.append(reference)
        return len(self.references)

    def grep(self, regexp):
        if regexp.search(self.title):
            return True

        if regexp.search(self.desc):
            return True

        for event in self.log_events:
            if regexp.search(event[3]):
                return True

        return False

    def make_id(self):
        strings = map(str, [self.creation_time, random.random(),
                            self.reporter, self.title, self.desc])
        return hashlib.sha1("\n".join(strings)).hexdigest()

    def __cmp__(self, other):
        if hasattr(other, "status"):
            val = cmp(SORT[self.status], SORT[other.status])
            if val:
                return val

        if hasattr(other, "creation_time"):
            return cmp(self.creation_time, other.creation_time)

        return 0

    def __repr__(self):
        return "<Issue: %s>" % self.id


class Config(DitzObject):
    yaml_tag = DitzObject.ditz_tag + '/config'
    filename = ".ditz-config"
    attributes = ["name", "email", "issue_dir"]

    def __init__(self, name, email, issue_dir="issues"):
        self.name = name
        self.email = email
        self.issue_dir = issue_dir

    @property
    def username(self):
        return "%s <%s>" % (self.name, self.email)

    def __repr__(self):
        return "<Config: %s>" % self.name


def write_yaml(item, fp=None, parent=None, level=0):
    """
    Write YAML data to a stream.
    """

    # This should be using PyYAML's write function, but I can't get it to
    # write things the way I want -- i.e., reproduce the format that
    # rubyditz uses.  And datetimes don't get written in a format that
    # rubyditz can read back again.  Sigh.

    # This function is ghastly and evil, and shouldn't exist.  But it gets
    # the job of roundtripping done.

    if fp is None:
        fp = sys.stdout

    value = unicode(item)
    tag = getattr(item, "yaml_tag", None)
    indent = "  "

    if tag:
        fp.write(tag + " \n")
        seenref = False
        for attr in item.attributes:
            if attr == "id" and not seenref:
                fp.write("\n")

            fp.write(indent * (level - 1))
            fp.write("%s: " % attr)
            obj = getattr(item, attr)

            if obj and attr == "references":
                seenref = True

            write_yaml(obj, fp, item, level + 1)
    elif isinstance(item, dict):
        if parent:
            fp.write("\n")

        for key, obj in sorted(item.items()):
            fp.write(indent * (level - 1))
            fp.write("%s: " % key)
            write_yaml(obj, fp, item, level + 1)
    elif isinstance(item, list):
        if len(item) > 0:
            newline = doindent = not isinstance(parent, list)

            # This is a gross hack.  Look away now.
            if isinstance(parent, DitzObject) and level > 1:
                level -= 1

            if newline:
                fp.write("\n")

            for obj in item:
                if doindent:
                    fp.write(indent * (level - 1))
                else:
                    doindent = True

                fp.write("- ")
                write_yaml(obj, fp, item, level + 1)
        else:
            fp.write("[]\n")
    elif isinstance(item, datetime):
        fp.write("%s Z\n" % str(item))
    elif "\n" in value:
        fp.write("|-\n")

        if isinstance(parent, list) and level > 1:
            level -= 1

        for line in value.split("\n"):
            fp.write(indent * level)
            fp.write(line + "\n")
    elif item is None:
        fp.write("\n")
    else:
        quote = False
        if not value:
            quote = True
        elif value[0] in '"{}[]':
            quote = True
        elif value[0] != ':' and ":" in value:
            quote = True
        else:
            try:
                float(value)
                quote = True
            except ValueError:
                pass

        if quote:
            value = '"' + value.replace('"', r'\"') + '"'

        fp.write(value + "\n")


def find_config(basedir=".", error=False):
    """
    Return a Ditz config file in or above a base directory.
    """

    basedir = os.path.realpath(basedir)
    filename = Config.filename
    curdir = basedir

    while True:
        path = os.path.join(curdir, filename)
        if os.path.exists(path):
            return path

        pardir = os.path.split(curdir)[0]
        if pardir == curdir:
            if error:
                raise DitzError("can't find '%s' in or above '%s'"
                                % (filename, basedir))
            else:
                return None

        curdir = pardir
